// noinspection JSCheckFunctionSignatures

import { writable } from 'svelte/store'
import { backendAddressAndPort, queryMode } from './constants.js'

const getRandomFromArray = array => array[Math.floor(Math.random() * array.length)]

let mockOffset = 1;

const addMetadata = mockItem => ({
  value: mockItem,
  magicByte: getRandomFromArray([0, 1, 2]),
  attributes: 0,
  timestamp: new Date().getTime(),
  offset: mockOffset++,
  key: null,
})

const mockItems = [
  {
    "eventType": "Television",
    "broughtToYouBy": "AMC",
    "title": "Breaking Bad"
  },
  {
    "eventType": "Television",
    "broughtToYouBy": "CBS",
    "title": "Breaking Bad Bang Theory"
  },
  {
    "eventType": "Movie",
    "broughtToYouBy": "20th Century Fox",
    "title": "Star Wars 2: The Star Warsening, brought to you by the Big Stink Corporation, a division of PepsiCo and somehow also Disney."
  },
].map(addMetadata)

export const testMode = false

export const state = writable({
  items: [],
  itemCount: undefined,
  matchCount: undefined,
  error: 'Connecting to WebSocket...',
  errorDetails: undefined,
})

const updateClearError = updater => state.update(s => {
  s.error = null
  s.errorDetails = null
  return updater(s)
})

let ws;
let itemLimit = Infinity
let disconnected = false
let filterFunc = (_message, _value, _json) => true

let runTestQuery = true
const testTimeout = 200
const testQuery = (mode) => {
  runTestQuery = true
  try {
    if (mode === queryMode.REAL_TIME) {
      const addItem = () => {
        const item = getRandomFromArray(mockItems)
        if (filterFunc(item, item.value, JSON.stringify(item))) {
          item.timestamp = new Date().getTime()
          state.update(s => ({
            ...s,
            items: [item, ...s.items].slice(0, itemLimit),
            itemCount: (s.itemCount || 0) + 1
          }))
        }
        if (runTestQuery) {
          setTimeout(addItem, testTimeout)
        }
      }
      setTimeout(addItem, testTimeout)
    } else {
      state.update(s => ({
        ...s,
        items: mockItems.filter(item => filterFunc(item, item.value, JSON.stringify(item))),
        itemCount: (s.itemCount || 0) + 1
      }))
    }
  } catch (e) {
    console.log('Caught an error:', e.toString())
    state.update(s => ({ ...s, error: e.toString() }))
  }
}

export const killQuery = async ({ }) => {
  // TODO
  if (testMode) {
    runTestQuery = false
    return
  }
  ws.send(JSON.stringify({
    mode: queryMode.KILL,
  }))
}

// noinspection JSUnusedGlobalSymbols
export const query = async ({ cluster, topic, mode, jsFilter, queryCode, maxItems, mutableObjects }) => {
  if (jsFilter) {
    filterFunc = new Function('message', 'value', 'json', queryCode)
  } else {
    filterFunc = () => true
  }
  itemLimit = maxItems

  updateClearError(s => ({ ...s, items: [], itemCount: 0 }))
  if (testMode) {
    testQuery(mode)
    return
  }

  if (disconnected) {
    connect()
    disconnected = false
  }

  ws.send(JSON.stringify({
    cluster,
    mode,
    topic,
    maxItems,
    immutable: !mutableObjects,
    searchCode: false //jsFilter && queryCode
  }))
}

export const connect = () => {
  if (ws?.readyState === WebSocket.OPEN) {
    return;
  }
  if (testMode) {
    updateClearError(s => ({ ...s, error: 'TEST MODE ENABLED' }))
    return
  }
  try {
    ws = new WebSocket(`ws://${backendAddressAndPort}`)
  } catch (e) {
    console.error(e.toString())
  }
  if (!ws) {
    updateClearError(s => ({ ...s, error: 'Unable to connect to websocket.' }))
    return
  }

  ws.addEventListener('open', () => {
    console.log('WebSocket opened')
  })

  ws.addEventListener('message', message => {
    if (message.data === 'CONNECTED') {
      updateClearError(s => ({...s}))
      return
    }
    let data;
    try {
      data = JSON.parse(message.data)
      switch (data?.type.toLowerCase()) {
        case 'complete':
          updateClearError(s => ({
            ...s,
            items: data.message
          }))
          break;
        case 'item_count':
          updateClearError(s => ({
            ...s,
            itemCount: console.log('state', s) || data.message
          }))
          break;
        case 'message':
          if (filterFunc(data.message, data.message.value, message.data)) {
            updateClearError(s => ({
              ...s,
              items: console.log('new item', data.message) || [data.message, ...s.items].slice(0, itemLimit),
              itemCount: (s.itemCount || 0) + 1
            }))
          }
          break;
      }
    } catch (e) {
      data = message.data
    }
  })

  ws.addEventListener('close', () => {
    disconnected = true
    state.update(s => ({ ...s, error: 'WebSocket connection closed.' }))
  })

  ws.addEventListener('error', () => {
    state.update(s => ({ ...s, error: 'WebSocket connection error.' }))
  })

  if (ws.readyState === ws.CLOSED || ws.readyState === ws.CLOSING) {
    state.update(s => ({ ...s, error: 'WebSocket connection closed.' }))
  }
}
